(*
  This file is part of Poussin.

  Poussin is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Poussin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Printf;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Varset;;


(* The beta reduction *)

let rec beta_red t bv def = 
  
  match t with
    | Var x -> Var x
    | Type -> Type
    | Lambda (x, t1,t2) -> Lambda (x, (beta_red t1 bv def), (beta_red t2 (x ++ bv) def))
    | Forall (x, t1,t2) -> Forall (x, (beta_red t1 bv def), (beta_red t2 (x ++ bv) def))
    | Ind (x, t1,t2) -> Ind (x, (beta_red t1 bv def), (beta_red_list t2 (x ++ bv) def))
    | Phi (x, n, t1,t2) -> 
        let t1' = (beta_red t1 bv def) in
        let (t1'_concl, t1'_lcv) = (forall_concl_list_var_type t1') in
          Phi (x, n, t1', (beta_red t2 (x ++ (bv +++ (context_to_setvar t1'_lcv))) def))        
    | Cons (n, t1) -> Cons (n, (beta_red t1 bv def))
    | App (t1,t2) -> (        
        match (beta_red t1 bv def) with
          | (Lambda (x, t11, t12)) ->
              beta_red 
                (*(rewrite_term_var_term (cf_term_listvar t12 (x ++ bv)) x (cf_term_listvar t2 (bv+++ (term_set_var (Lambda (x, t11, t12))))))*)
                (subst_var_term t12 x t2 bv)
                (bv+++ (term_set_var (Lambda (x, t11, t12)))) def
          | (Phi (x, n, t11, t12)) ->              
              let (type_concl, type_lcv) = (forall_nth_concl_list_var_type t11 n) in
                beta_red (
                  App(
                    concl_list_var_type_lambda
                      (*
                        (rewrite_term_var_term 
                        (cf_term_listvar t12 (x ++ bv))
                        x 
                        (cf_term_listvar (Phi (x, n, t11, t12)) (bv +++ (term_set_var (Phi (x, n, t11, t12)))))
                        )
                      *)
                      (subst_var_term t12 x (Phi (x, n, t11, t12)) bv)
                      type_lcv
                      ,
                    t2
                  )
                )
                  
                  (bv+++ (term_set_var (Phi (x, n, t11, t12))))
                  def
                  
          | t1' -> App (t1', (beta_red t2 bv def))
              (* here should be a rule for unfolding *)
      )
    | Match (t1, t2) -> (
        
        let t1' = (beta_red t1 bv def) in
          match (app_first_term_args t1') with
            | (Cons (n, Ind (x, t11,t12)), args) -> (
                let t2' = (beta_red_list t2 bv def) in
                  match (nth t2' n) with
                    | None -> Match (t1', t2')
                    | Some t' ->
                        beta_red (listargs_app t' args) bv def                
              )
            | _ -> Match (t1', t2)
                (* here should be a rule for unfolding *)

      )

    | MatchT (t1, t2, t3) -> (
        
        let t1' = (beta_red t1 bv def) in
        let t3' = (beta_red t3 bv def) in
          match (app_first_term_args t1') with
            | (Cons (n, Ind (x, t11,t12)), args) -> (
                let t2' = (beta_red_list t2 bv def) in
                  match (nth t2' n) with
                    | None -> MatchT (t1', t2', t3')
                    | Some t' ->
                        beta_red (listargs_app t' args) bv def               
              )
            | _ -> MatchT (t1', t2, t3')
                (* here should be a rule for unfolding *)
      )

and beta_red_list l bv def =
  match l with
    | [] -> []
    | hd :: tl ->
        (beta_red hd bv def) :: (beta_red_list tl bv def);;



(*************************)

let rec nomatch t bv =
  match t with
    | Var x -> true
    | Type -> true
    | Lambda (x, t1,t2) -> (nomatch  t1 (x ++ bv)) && (nomatch  t2 (x ++ bv))
    | Forall (x, t1,t2) -> (nomatch  t1 (x ++ bv)) && (nomatch  t2 (x ++ bv))
    | Ind (x, t1,t2) -> true
    | Phi (x, n, t1,t2) -> (nomatch  t1 (x ++ bv)) && 
        let (type_concl, type_lcv) = (forall_nth_concl_list_var_type t1 n) in           
          (nomatch  t2 (x ++ (bv +++ (context_to_setvar type_lcv))))
    | Cons (n, t1) -> (nomatch  t1 bv)
    | App (t1,t2) -> (nomatch  t1 bv) && (nomatch  t2 bv)
    | Match (t1, t2) -> (
        match t1 with
          | Var x -> if (VarSet.subset (VarSet.singleton x) bv) then (nomatch_list t2 bv) else false
          | _ -> false
      )
    | MatchT (t1, t2, t3) -> (
        match t1 with
          | Var x -> if (VarSet.subset (VarSet.singleton x) bv) then (nomatch_list t2 bv) else false
          | _ -> false
      )
and nomatch_list l bv =
  match l with
    | [] -> true
    | hd::tl -> (nomatch hd bv) && (nomatch_list tl bv);;


(*****)

(* a simplified beta reduction for the simpl tactic *)


let rec simpl_red t bv def = 
  (*print_term (listfold_term_term_var_fp t def); printf "\n\n"; flush stdout;*)
  match t with
    | Var x -> Var x
    | Type -> Type
    | Lambda (x, t1,t2) -> Lambda (x, (simpl_red t1 bv def), (simpl_red t2 (x ++ bv) def))
    | Forall (x, t1,t2) -> Forall (x, (simpl_red t1 bv def), (simpl_red t2 (x ++ bv) def))
    | Ind (x, t1,t2) -> Ind (x, (simpl_red t1 bv def), (simpl_red_list t2 (x ++ bv) def))
    | Phi (x, n, t1,t2) -> Phi (x, n, (simpl_red t1 bv def), t2)        
    | Cons (n, t1) -> Cons (n, (simpl_red t1 bv def))
    | Match (t1, t2) -> (
        
        let t1' = (simpl_red t1 bv def) in
          match (app_first_term_args t1') with
            | (Cons (n, Ind (x, t11,t12)), args) -> (
                match (nth t2 n) with
                  | None -> Match (t1', t2)
                  | Some t' ->
                      simpl_red (listargs_app t' args) bv def                
              )
            | _ -> Match (t1', t2)
                (* here should be a rule for unfolding *)
      )

    | MatchT (t1, t2, t3) -> (
        
        let t1' = (simpl_red t1 bv def) in
        let t3' = (simpl_red t3 bv def) in
          match (app_first_term_args t1') with
            | (Cons (n, Ind (x, t11,t12)), args) -> (
                match (nth t2 n) with
                  | None -> MatchT (t1', t2, t3')
                  | Some t' ->
                      simpl_red (listargs_app t' args) bv def               
              )
            | _ -> MatchT (t1', t2, t3')
                (* here should be a rule for unfolding *)
      )
    | App (t1,t2) -> (        
        let (fct, args) = (app_first_term_args t) in
        let (fct', args') = (simpl_red fct bv def, simpl_red_list args bv def) in
        let (fct'', args'') = (simpl_red_app fct' args' bv def) in
          listargs_app fct'' args''
      )


and simpl_red_list l bv def =
  match l with
    | [] -> []
    | hd :: tl -> (simpl_red hd bv def) :: (simpl_red_list tl bv def)

and simpl_red_app fct args bv def =
  match args with
    | [] -> (fct, [])
    | hd :: tl -> (
        match fct with
          | Lambda (x, t11, t12) ->
              simpl_red_app 
                
                (simpl_red 
                    (*
                      (rewrite_term_var_term 
                      (cf_term_listvar t12 (term_set_var hd)) 
                      x 
                      (cf_term_listvar hd (bv +++ (term_set_var (Lambda (x, t11, t12)))))
                      )
                    *)
                    (subst_var_term t12 x hd bv)
                    (bv +++ (term_set_var (Lambda (x, t11, t12)))  +++ (term_set_var hd)) 
                    def)

                tl
                
                bv
                
                def
                
          | Phi (x, n, t11, t12) ->
              let (type_concl, type_lcv) = (forall_nth_concl_list_var_type t11 n) in
              let fct' = ( 
                
                concl_list_var_type_lambda
                  (*
                    (
                    rewrite_term_var_term 
                    (cf_term_listvar t12 (term_set_var hd)) 
                    x 
                    (cf_term_listvar (Phi (x, n, t11, t12)) (bv +++ (term_set_var (Phi (x, n, t11, t12))) +++ (term_set_var hd)))
                    )
                  *)
                  (subst_var_term t12 x (Phi (x, n, t11, t12)) bv)
                  (type_lcv)
                  
                  
              ) in 
              let (t1, t2) = (simpl_red_app fct' args (bv +++ (term_set_var (Phi (x, n, t11, t12))) +++ (term_set_var hd)) def) in
                if (nomatch t1 VarSet.empty) then (t1, t2) else (
                  
                  (*print_term t1; printf "\n\n";*)
                  (fct, args)
                )
                  
          | _ -> (fct, args)                
              (* here should be a rule for unfolding *)
      );;

