(*
  This file is part of Poussin.

  Poussin is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Poussin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Printf;;
open Lexing;;
open List;;
open String;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Alpharel;;
open Betared;;
open Termequiv;;
open Typechecking
open Unfold;;
open Varset;;
open Kernel;;
open Unification;;

type proof_tree =
    (((context * string * term) list) * (string * term * term)) option;;  

type int_state = {mutable n: int; mutable def: context list; mutable subgoals: int; mutable pt: proof_tree; mutable ctxt: context; mutable lemmas_def: context};;

let state : int_state = {n = 0; def = []; subgoals = 0; pt = None; ctxt = []; lemmas_def = []};;

let rec listconstructors2def l n name =
  match l with 
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, Cons (n,Var (name)))::(listconstructors2def tl (n+1) name);;

let rec append_foralls_constructors lc ll =
  match lc with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, concl_list_var_type_forall hd2 ll) :: (append_foralls_constructors tl ll);;

let rec listdef2def l =
  match l with
    | [] -> []
    | hd :: tl -> 
        hd @ (listdef2def tl);;

let proof_mode () =
  match (state.pt) with
    | None -> false
    | Some _ -> true;;

let rec print_subgoals l =
  match l with
    | [] -> printf "";
    | (hd1, hd2, hd3) ::tl ->
        printf "%s: " hd2; print_term hd3; printf "\n";
        print_subgoals tl;;

let rec print_proof l g =
  print_subgoals l; 
  let (hd1, hd2, hd3) = g in (
      printf "Lemma %s : " hd1; print_term hd3; printf " := ";
      print_term hd2;
      printf "\n\n";
    );;

let rec context_list_var c =
  match c with
    | [] -> []
    | (hd1,hd2)::tl ->
        hd1::(context_list_var tl);;

let rec def_list_var d =
  match d with
    | [] -> []
    | hd :: tl ->
        (context_list_var hd) @ (def_list_var tl);;

let rec def_set_var d =
  match d with
    | [] -> VarSet.empty
    | hd :: tl ->
        (context_set_var hd) +++ (def_set_var tl);;

(********************)

(* Automatic generation of induction principles *)


(* generation of induction type *)

let rec ind_princ_cons_ind_hyp lqv ind_name consnum varnum pred ind_args ind_type_hyp =
  match lqv with
    | [] -> []
    | (hd1, hd2) :: tl ->
        let tl' = (ind_princ_cons_ind_hyp tl ind_name consnum (varnum + 1) pred ind_args ind_type_hyp) in
        let (fct, args) = (app_first_term_args hd2) in
        let pred_args = (
          match (unification_list (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp))) args VarSet.empty) with
            | Mgu s -> 
                (*print_context (readable_context_fp (comp_subs s) state.def); printf "\n\n";*)
                apply_subst_listterm (list_name2list_var (list_proj1 (ind_type_hyp))) (comp_subs s) VarSet.empty
            | _ -> (list_name2list_var (list_proj1 (ind_type_hyp)))          
        ) in
          if (fct = Var ind_name) then
            (
              (concat "" ("Pred" :: "indind" :: (string_of_int consnum) :: (string_of_int varnum) :: [])),
              (listargs_app (Var pred) (pred_args @ (Var hd1)::[]))
            ) :: tl' 
          else 
            tl';;

let rec ind_princ_cons lcons ind_args pred cons_num ind_name ind_type_hyp ind_def =
  match lcons with
    | [] -> []
    | hd :: tl ->
        let tl' = (ind_princ_cons tl ind_args pred (cons_num + 1) ind_name ind_type_hyp ind_def) in
        let (hd_concl, hd_hyp) = (forall_concl_list_var_type hd) in
        let (hd_concl_fct, hd_concl_args) = (app_first_term_args hd_concl) in
        let pred_args = (
          match (unification_list (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp))) hd_concl_args VarSet.empty) with
            | Mgu s -> 
                (*print_context (readable_context_fp (comp_subs s) state.def); printf "\n\n";*)
                apply_subst_listterm (list_name2list_var (list_proj1 (ind_type_hyp))) (comp_subs s) VarSet.empty
            | _ -> (list_name2list_var (list_proj1 (ind_type_hyp)))
        ) in
        let ind_hyp = (ind_princ_cons_ind_hyp hd_hyp ind_name cons_num 0 pred ind_args ind_type_hyp) in

          ( concat "" (pred :: "ind" :: (string_of_int cons_num) :: []),
            (concl_list_var_type_forall
               ( App
                   (
                     listargs_app (Var pred) pred_args,
                     listargs_app (Cons (cons_num,(Var ind_name))) ((list_name2list_var (list_proj1 (ind_args @ hd_hyp))))                  
                   )
               )
               
               (hd_hyp @ ind_hyp)
            )
          ) :: tl';;

let ind_princ_type ind_name ind_args ind_type ind_cons pred =
  let (ind_type_concl, ind_type_hyp) = (forall_concl_list_var_type ind_type) in
  (*let ind = (Ind(ind_name, concl_list_var_type_forall ind_type ind_args, list_concl_list_var_type_forall ind_cons ind_args)) in*)
  let ind_princ_hyp = (ind_princ_cons ind_cons ind_args pred 0 ind_name ind_type_hyp (Var ind_name)) in
    (concl_list_var_type_forall
       
       (listargs_app (Var pred) ((list_name2list_var (list_proj1 ind_type_hyp) @ (Var "x")::[])))
       
       (ind_args @ (pred, concl_list_var_type_forall (Forall ("x", listargs_app (Var ind_name) (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp))), Type)) ind_type_hyp) :: ind_princ_hyp @ ind_type_hyp @ ("x", listargs_app (Var ind_name) (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp)))) :: [])
    );;


(* generation of induction principle *)

let rec ind_princ_dest_rec_hyp lqv ind_name ind_type_hyp ind_args =
  match lqv with
    | [] -> []
    | (hd1, hd2) :: tl ->
        let tl' = (ind_princ_dest_rec_hyp tl ind_name ind_type_hyp ind_args) in
        let (fct, args) = (app_first_term_args hd2) in
        let pred_args = (
          match (unification_list (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp))) args VarSet.empty) with
            | Mgu s -> 
                (*print_context (readable_context_fp (comp_subs s) state.def); printf "\n\n";*)
                apply_subst_listterm (list_name2list_var (list_proj1 (ind_type_hyp))) (comp_subs s) VarSet.empty
            | _ -> (list_name2list_var (list_proj1 (ind_type_hyp)))
        ) in
          if (fct = Var ind_name) then
            (listargs_app
               (Var (concat "" (ind_name :: "induc" :: [])))
               (pred_args @ (Var hd1)::[])
            ) :: tl' 
          else 
            tl';;

let rec ind_princ_dest lcons ind_args ind_name cons_num pred ind_type_hyp =
  match lcons with
    | [] -> []
    | hd :: tl ->
        let tl' = (ind_princ_dest tl ind_args ind_name (cons_num + 1) pred ind_type_hyp) in
        let (cons_concl, cons_hyp) = (forall_concl_list_var_type hd) in
          (concl_list_var_type_lambda
             (
               listargs_app
                 (Var (concat "" (pred :: "ind"  :: (string_of_int cons_num) :: [])))
                 
                 (
                   (list_name2list_var (list_proj1 cons_hyp)) @
                     (ind_princ_dest_rec_hyp cons_hyp ind_name ind_type_hyp ind_args)
                 )
             )
             (ind_args @ cons_hyp)
          ) :: tl';;

let ind_princ_pf ind_name ind_args ind_type ind_cons pred =
  let (ind_type_concl, ind_type_hyp) = (forall_concl_list_var_type ind_type) in
  let ind = (Ind(ind_name, concl_list_var_type_forall ind_type ind_args, list_concl_list_var_type_forall ind_cons ind_args)) in
  let ind_princ_hyp = (ind_princ_cons ind_cons ind_args pred 0 ind_name ind_type_hyp ind) in
  let dest = (ind_princ_dest ind_cons ind_args ind_name 0 pred ind_type_hyp) in
  let cur_ind_type = listargs_app (Var ind_name) (list_name2list_var (list_proj1 (ind_args @ ind_type_hyp))) in
    (concl_list_var_type_lambda
       ( Phi 
           (
             (concat "" (ind_name :: "induc" :: [])) , 
             List.length ind_type_hyp + 1,
             (
               concl_list_var_type_forall (Forall ("x", cur_ind_type, (listargs_app (Var pred) ((list_name2list_var (list_proj1 ind_type_hyp) @ (Var "x")::[])))))
                 ind_type_hyp
             ),
             (
               ( MatchT 
                   (
                     (Var "x"),                        
                     (dest),
                     (listargs_app (Var pred) ((list_name2list_var (list_proj1 ind_type_hyp) @ (Var "x")::[])))
                   )
               )                
             )
           )
       )
       
       (
         ind_args @  
           (pred, concl_list_var_type_forall (Forall ("x", cur_ind_type, Type)) ind_type_hyp) :: ind_princ_hyp
       )
    );;

(*****)

let add_inductive_woa name ty cons =

  let poussin_term = Ind (name, ty, list_proj2 cons) in
  let def = ((listconstructors2def cons 0 name) @ (name, poussin_term)::[]) in 

    if (typecheck 
          state.ctxt
          poussin_term 
          ty 
          (listdef2def state.def)
          VarSet.empty 
       ) then (          
      
      state.def <- def::state.def;
      state.ctxt <- cons @ (name, ty)::state.ctxt ;
      
      printf "Inductive %s is defined.\n\n" name;  
      
      let ty' = (ind_princ_type name [] ty (list_proj2 cons) "Pred") in
      let te' = (ind_princ_pf name [] ty (list_proj2 cons) "Pred") in
        if (typecheck 
              state.ctxt
              te'
              ty'
              (listdef2def (def::state.def))
              VarSet.empty
           ) then (
          
          printf "Induction principle %s is defined.\n\n" (concat "" (name :: "induc" :: []));              
          (*
            state.lemmas <- (concat "" (name :: "induc" :: []), (readable_term_fp ty' (def::state.def)))::state.lemmas;
            state.lemmas_def <- (concat "" (name :: "induc" :: []), (readable_term_fp te' (def::state.def)))::state.lemmas_def;
          *)
          
          state.ctxt <- (concat "" (name :: "induc" :: []), (readable_term_fp ty' (def::state.def)))::state.ctxt;
          state.def <- ((concat "" (name :: "induc" :: []), (readable_term_fp te' (def::state.def)))::[])::state.def;

          ) else (
          
          printf "Proposition type for induction principle :\n";
          print_term (readable_term_fp ty' (def::state.def)); printf "\n\n";
          printf "Proposition term for induction principle :\n";
          print_term te'; printf "\n\n";
            
        )
          
    ) else (       
      
      printf "Error: Inductive %s does not type!\n\n" name;
      
    ); flush stdout;;


let add_inductive_wa name args ty cons =

  let poussin_term = (      
    Ind (name, concl_list_var_type_forall ty args, list_proj2 (append_foralls_constructors cons args))
  ) in
  let def = ((listconstructors2def (append_foralls_constructors cons args) 0 name) @ (name, poussin_term)::[]) in (      

      (*printf "term: "; print_term poussin_term; printf "\n\n";*)
      
      if (typecheck 
            state.ctxt
            poussin_term 
            (concl_list_var_type_forall ty args)
            (listdef2def state.def)
            VarSet.empty
         ) then (         
        
        printf "Inductive %s is defined.\n\n" name;  
        
        state.def <- def::state.def;
        state.ctxt <- (append_foralls_constructors cons args) @ (name, concl_list_var_type_forall ty args)::state.ctxt ;

        let ty' = (ind_princ_type name args ty (list_proj2 cons) "Pred") in
        let te' = (ind_princ_pf name args ty (list_proj2 cons) "Pred") in
          if (typecheck 
                state.ctxt
                te'
                ty'
                (listdef2def (def::state.def))
                VarSet.empty
             ) then (
            
            printf "Induction principle %s is defined.\n\n" (concat "" (name :: "induc" :: []));  
            (*
            state.lemmas <- (concat "" (name :: "induc" :: []), (readable_term_fp ty' (def::state.def)))::state.lemmas;
            state.lemmas_def <- (concat "" (name :: "induc" :: []), (readable_term_fp te' (def::state.def)))::state.lemmas_def;
            *)

          state.ctxt <- (concat "" (name :: "induc" :: []), (readable_term_fp ty' (def::state.def)))::state.ctxt;
          state.def <- ((concat "" (name :: "induc" :: []), (readable_term_fp te' (def::state.def)))::[])::state.def;

              
          ) else (
            
            printf "Proposition type for induction principle :\n";
            print_term (readable_term_fp ty' (def::state.def)); printf "\n\n";
            printf "Proposition term for induction principle :\n";
            print_term (readable_term_fp te' (def::state.def)); printf "\n\n";
            

          )

      ) else (       

        printf "Error: Inductive %s does not type!\n\n" name;

      )        
    ); flush stdout;;


let add_recursive name args ty def =

  let poussin_term = (      
    
    Phi (name, List.length args , concl_list_var_type_forall ty args, def)

  ) in (

      if (typecheck 
            state.ctxt
            poussin_term
            (concl_list_var_type_forall ty args)
            (listdef2def state.def)
            VarSet.empty
         ) then (          
        
        printf "Recursive %s is defined.\n\n" name; 

        state.def <- ((name, poussin_term)::[])::state.def;
        state.ctxt <- (name, concl_list_var_type_forall ty args) :: state.ctxt;

        
      ) else (       
        printf "Error: Recursive %s does not type!\n\n" name;
      )        
    ); flush stdout;;



let add_definition_woa name ty def =
  
  if (typecheck 
        state.ctxt         
        def
        ty
        (listdef2def state.def)
        VarSet.empty
     ) then (          
    
    state.def <- ((name,def)::[])::state.def;
    state.ctxt <- (name,ty)::state.ctxt;

    printf "Definition %s is defined.\n\n" name; 
    
  ) else (       
    
    printf "Error: Definition %s does not type!\n\n" name;
    
  ); flush stdout;;



let add_definition_wa name args ty def =
  
  let poussin_term = (concl_list_var_type_lambda def args) in (
      
      if (typecheck 
            state.ctxt           
            poussin_term
            (concl_list_var_type_forall ty args)
            (listdef2def state.def)
            VarSet.empty
         ) then (          
        
        state.def <- ((name,poussin_term)::[])::state.def;
        state.ctxt <- (name,(concl_list_var_type_forall ty args))::state.ctxt;

        printf "Definition %s is defined.\n\n" name; 
        
      ) else (       
        
        printf "Error: Definition %s does not type!\n\n" name;
        
      )
    ); flush stdout;;


