(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Def;;
open Printf;;
open Freshness;;
open Pprinter;;
open Listutil;;
open Varset;;

(* returns the list of arguments of an application *)

let rec app_list_args t =
  match t with
    | App (t1, t2) ->
        ((app_list_args t1) @ (t2::[]))
    | _ -> [];;

(* return the function of an application *)

let rec app_first_term t =
  match t with
    | App (t1, t2) ->
        (app_first_term t1)
    | _ -> t;;

(* returns a couple of function and its arguments *)
let rec app_first_term_args t =
  (app_first_term t, app_list_args t);;


(* returns the application of the term t with arguments args *)

let rec listargs_app t args =
  match args with
    | [] -> t
    | hd::tl ->
        listargs_app (App (t, hd)) tl;;


(* returns the list of contiguously V quantified variables with their types *)

let rec forall_list_var_type t =
  match t with
    | Forall (x,t1,t2) -> 
        (x,t1)::(forall_list_var_type t2)
    | _ -> [];;

(* returns the term without the first V quantifications *)

let rec forall_concl t =
  match t with
    | Forall (x,t1,t2) -> 
        (forall_concl t2)
    | _ -> t;;

(* returns the couples of both previous functions *)

let forall_concl_list_var_type t =
  (forall_concl t, forall_list_var_type t);;

(* returns a term c V quantified by the list of variable/type l *)

let rec concl_list_var_type_forall c l =
  match l with
    | [] -> c
    | (hd1,hd2) :: tl ->
        (Forall (hd1,hd2, (concl_list_var_type_forall c tl)));;


let rec list_concl_list_var_type_forall lc l =
  match lc with
    | [] -> []
    | hd :: tl ->
        (concl_list_var_type_forall hd l) :: (list_concl_list_var_type_forall tl l);;

(* returns the list of the nth first contiguously V quantified variables with their types *)

let rec forall_nth_list_var_type t n =
  match n with
    | 0 -> []
    | _ ->  
        match t with
          | Forall (x,t1,t2) -> 
              (x,t1)::(forall_nth_list_var_type t2 (n-1))
          | _ -> [];;

(* returns the term without the first V quantifications *)

let rec forall_nth_concl t n =
  match n with
    | 0 -> t
    | _ ->
        match t with
          | Forall (x,t1,t2) -> 
              (forall_nth_concl t2 (n-1))
          | _ -> t;;

(* returns the couples of both previous functions *)

let forall_nth_concl_list_var_type t n =
  (forall_nth_concl t n, forall_nth_list_var_type t n);;

(* returns the list of the nth first contiguously \ quantified variables with their types *)

let rec lambda_list_var_type t n =
  match n with
    | 0 -> []
    | _ ->
        match t with
          | Lambda (x,t1,t2) -> 
              (x,t1)::(lambda_list_var_type t2 (n - 1))
          | _ -> [];;

(* returns the term without the first \ quantifications *)

let rec lambda_concl t n =
  match n with
    | 0 -> t
    | _ ->
        match t with
          | Lambda (x,t1,t2) -> 
              (lambda_concl t2 (n - 1))
          | _ -> t;;

(* returns the couples of both previous functions *)

let lambda_concl_list_var_type t n =
  (lambda_concl t n, lambda_list_var_type t n);;

(* returns a term c V quantified by the list of variable/type l *)

let rec concl_list_var_type_lambda c l =
  match l with
    | [] -> c
    | (hd1,hd2) :: tl ->
        (Lambda (hd1,hd2, (concl_list_var_type_lambda c tl)));;

(* find the type of a variable in a context *)

let rec var_type_in_context x ctxt =
  match ctxt with
    | [] -> None
    | (hd1, hd2)::tl ->
        if (hd1 = x) then Some hd2 else var_type_in_context x tl;;

(* Transform a context into a list of var *)

let rec context_to_listvar ctxt =
  match ctxt with
    | [] -> []
    | (hd1, hd2) :: tl -> hd1 :: context_to_listvar tl;;

let rec context_to_setvar ctxt =
  match ctxt with
    | [] -> VarSet.empty
    | (hd1, hd2) :: tl -> hd1 ++ context_to_setvar tl;;

let rec context_to_listtermofvar ctxt =
  match ctxt with
    | [] -> []
    | (hd1, hd2) :: tl -> (Var hd1) :: context_to_listtermofvar tl;;


let rec list_name2list_var l =
  match l with
    | [] -> []
    | hd :: tl ->
        (Var hd) :: (list_name2list_var tl);;



(* returns the set of used variables by a term *)
let rec term_set_var t =
  match t with
    | Var v -> 
        VarSet.singleton v
    | Lambda (s, t1, t2) ->
         s ++ ((term_set_var t1) +++ (term_set_var t2))
    | Forall (s, t1, t2) ->
        s ++ ((term_set_var t1) +++ (term_set_var t2))
    | Phi (s, n, t1, t2) ->
        s ++ ((term_set_var t1) +++ (term_set_var t2))
    | Ind (s, t1, l) ->
        s ++ ((term_set_var t1) +++ (listterm_set_var l))
    | App (t1, t2) ->
        (term_set_var t1) +++ (term_set_var t2)
    | Match (t1, l) ->
        (term_set_var t1) +++ (listterm_set_var l)
    | MatchT (t1, l, t2) ->
        (listterm_set_var l) +++ (term_set_var t1) +++ (term_set_var t2)
    | Cons (n, t1) ->
        (term_set_var t1)
    | Type -> VarSet.empty
and listterm_set_var l =
  match l with
    | [] -> VarSet.empty
    | hd :: tl ->
        (term_set_var hd) +++ (listterm_set_var tl);;

let rec context_set_var c =
  match c with
    | [] -> VarSet.empty;
    | (hd1,hd2) :: tl ->
        (hd1 ++ (term_set_var hd2)) +++ (context_set_var tl);;

(* returns the list of free variables *)
let rec term_fv t bv =
  match t with
    | Var v -> 
        if (VarSet.subset (VarSet.singleton v) bv) then VarSet.empty else (VarSet.singleton v)
    | Lambda (s, t1, t2) ->
        (term_fv t1 (s ++ bv)) +++ (term_fv t2 (s ++ bv))
    | Forall (s, t1, t2) ->
        (term_fv t1 (s ++ bv)) +++ (term_fv t2 (s ++ bv))
    | Phi (s, n, t1, t2) ->
        (term_fv t1 (s ++ bv)) +++ (term_fv t2 (s ++ bv))
    | Ind (s, t1, l) ->
        (term_fv t1 (s ++ bv)) +++ (listterm_fv l (s ++ bv))
    | App (t1, t2) ->
        (term_fv t1 bv) +++ (term_fv t2 bv)
    | Match (t1, l) ->
        (term_fv t1 bv) +++ (listterm_fv l bv)
    | MatchT (t1, l, t2) ->
        (term_fv t1 bv) +++ (listterm_fv l bv)
    | Cons (n, t1) ->
        (term_fv t1 bv)
    | Type -> VarSet.empty
and listterm_fv lt bv =
  match lt with
    | [] -> VarSet.empty
    | hd :: tl ->
        (term_fv hd bv) +++ (listterm_fv tl bv);;


(* rewriting of a variable x by a variable y in the term t (smooth) *)
let rec rewrite_term_var_var t x y =
  match t with
    | Var v -> 
        if (v = x) then (Var y) else (Var v)
    | Lambda (s, t1, t2) ->
        if (s = x) then
          (Lambda (y, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y)) else
            (Lambda (s, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y))
    | Forall (s, t1, t2) ->
        if (s = x) then
          (Forall (y, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y)) else
            (Forall (s, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y))
    | Phi (s, n, t1, t2) ->
        if (s = x) then
          (Phi (y, n, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y)) else
            (Phi (s, n, rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y))
    | Ind (s, t1, l) ->
        if (s = x) then
          (Ind (y, rewrite_term_var_var t1 x y, rewrite_listterm_var_var l x y)) else
            (Ind (s, rewrite_term_var_var t1 x y, rewrite_listterm_var_var l x y))
    | App (t1, t2) ->
        App (rewrite_term_var_var t1 x y, rewrite_term_var_var t2 x y)
    | Match (t1, l) ->
        Match (rewrite_term_var_var t1 x y, rewrite_listterm_var_var l x y)
    | MatchT (t1, l, t2) ->
        MatchT (rewrite_term_var_var t1 x y, rewrite_listterm_var_var l x y, rewrite_term_var_var t2 x y)
    | Cons (n, l) ->
        Cons (n, rewrite_term_var_var l x y)
    | Type -> Type
and rewrite_listterm_var_var l x y =
  match l with
    | [] -> []
    | hd :: tl ->
        (rewrite_term_var_var hd x y) :: (rewrite_listterm_var_var tl x y);;

(* rewriting of a variable x by a term y in the term t (quantifier sensible) *)
let rec rewrite_term_var_term t x y =
  match t with
    | Var v -> 
        if (v = x) then y else (Var v)
    | Lambda (s, t1, t2) ->
        if (s = x) then
          t else
            (Lambda (s, rewrite_term_var_term t1 x y, rewrite_term_var_term t2 x y))
    | Forall (s, t1, t2) ->
        if (s = x) then
           t else
            (Forall (s, rewrite_term_var_term t1 x y, rewrite_term_var_term t2 x y))
    | Phi (s, n, t1, t2) ->

        if (s = x) then
          t else
          (Phi (s, n, rewrite_term_var_term t1 x y, 
               if (VarSet.subset (VarSet.singleton x) (context_to_setvar (forall_nth_list_var_type t1 n))) then t2 else (rewrite_term_var_term t2 x y)
          )
          )

    | Ind (s, t1, l) ->
        if (s = x) then
          t else
            (Ind (s, rewrite_term_var_term t1 x y, rewrite_listterm_var_term l x y))
    | App (t1, t2) ->
        App (rewrite_term_var_term t1 x y, rewrite_term_var_term t2 x y)
    | Match (t1, l) ->
        Match (rewrite_term_var_term t1 x y, rewrite_listterm_var_term l x y)
    | MatchT (t1, l, t2) ->
        MatchT (rewrite_term_var_term t1 x y, rewrite_listterm_var_term l x y, rewrite_term_var_term t2 x y)
    | Cons (n, l) ->
        Cons (n, rewrite_term_var_term l x y)
    | Type -> Type
and rewrite_listterm_var_term l x y =
  match l with
    | [] -> []
    | hd :: tl ->
        (rewrite_term_var_term hd x y) :: (rewrite_listterm_var_term tl x y);;

let rec listrewrite_term_var_term t l =
  match l with
    | [] -> t
    | (hd1, hd2) :: tl ->
        listrewrite_term_var_term (rewrite_term_var_term t hd1 hd2) tl;;

let rec listrewrite_listterm_var_term lt l =
  match lt with
    | [] -> []
    | hd :: tl ->
        (listrewrite_term_var_term hd l) :: (listrewrite_listterm_var_term tl l);;

let rec list_listrewrite_term_var_term t l =
  match l with
    | [] -> t
    | hd::tl ->
        list_listrewrite_term_var_term (listrewrite_term_var_term t hd) tl;;



let rec listrewrite_term_vare_term t l =
  match l with
    | [] -> t
    | (Var x, te)::tl ->
        listrewrite_term_vare_term (rewrite_term_var_term t x te) tl
    | (_ :: tl) -> listrewrite_term_vare_term t tl ;;

let rec listrewrite_listterm_vare_term lt l =
  match lt with
    | [] -> []
    | hd :: tl ->
        (listrewrite_term_vare_term hd l) :: (listrewrite_listterm_vare_term tl l);;

(*[ pas sur la ... ]*)

let rec listrewrite_context_vare_term c l =
  match c with 
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, listrewrite_term_vare_term hd2 l) :: (listrewrite_context_vare_term tl l);;


let rec rewrite_context_var_term l x y =
  match l with
    | [] -> []
    | (hd1, hd2) :: tl ->
        if (x = hd1) then
          (hd1, hd2) :: (rewrite_context_var_term tl x y) else
            (hd1, rewrite_term_var_term hd2 x y) :: (rewrite_context_var_term tl x y);;

let rec listrewrite_context_var_term c l=
  match l with
    | [] -> c
    | (hd1, hd2) :: tl ->
        listrewrite_context_var_term (rewrite_context_var_term c hd1 hd2) tl;;


(* substitution *)

let rec apply_subs_var x s =
  match s with
    | [] -> Var x
    | (v, t) :: tl ->
        if (x = v) then t else apply_subs_var x tl;;

let rec apply_subs_term t s bv =
  match t with
    | Var x -> 
        if (VarSet.subset (VarSet.singleton x) bv) then (Var x) else (apply_subs_var x s)
    | Lambda (x, t1, t2) ->
        Lambda (x, (apply_subs_term t1 s (x ++ bv)), (apply_subs_term t2 s (x ++ bv)))
    | Forall (x, t1, t2) ->
        Forall (x, (apply_subs_term t1 s ( x ++ bv)), (apply_subs_term t2 s ( x ++ bv)))
    | Phi (x, n, t1, t2) ->
        Phi (x, n, (apply_subs_term t1 s ( x ++ bv)), (apply_subs_term t2 s ( x ++ (bv +++ (context_to_setvar (forall_nth_list_var_type t1 n))))))
    | Ind (x, t1, t2) ->
        Ind (x, (apply_subs_term t1 s ( x ++ bv)), (apply_subs_listterm t2 s ( x ++ bv)))
    | App (t1, t2) ->
        App ((apply_subs_term t1 s bv), (apply_subs_term t2 s bv))
    | Match (t1, l1) ->
        Match ((apply_subs_term t1 s bv), (apply_subs_listterm l1 s bv))
    | MatchT (t1, l1, t2) ->
        MatchT ((apply_subs_term t1 s bv), (apply_subs_listterm l1 s bv),(apply_subs_term t2 s bv))
    | Cons (n, l1) ->
        Cons (n, (apply_subs_term l1 s bv))
    | Type -> Type
and apply_subs_listterm lt s bv =
  match lt with
    | [] -> []
    | hd :: tl ->
        (apply_subs_term hd s bv)::(apply_subs_listterm tl s bv);;

(* transforms a term in a canonical form:

   all quantified variables are different

*)

let rec cf_term' t l =
  match t with
    | Lambda (x, t1, t2) ->
        if (VarSet.subset (VarSet.singleton x) l) then
          let y = (fresh_name_list_name l x) in
          let (t1', l1') = (cf_term' t1 (y ++ l)) in
          let (t2', l2') = (cf_term' t2 l1') in
            (Lambda (y, rewrite_term_var_var t1' x y, rewrite_term_var_var t2' x y), l2')
        else
          let (t1', l1') = (cf_term' t1 (x ++ l)) in
          let (t2', l2') = (cf_term' t2 l1') in
            (Lambda (x, t1', t2'), l2')
    | Forall (x, t1, t2) ->
        if (VarSet.subset (VarSet.singleton x) l) then
          let y = (fresh_name_list_name l x) in
          let (t1', l1') = (cf_term' t1 (y ++ l)) in
          let (t2', l2') = (cf_term' t2 l1') in
            (Forall (y, rewrite_term_var_var t1' x y, rewrite_term_var_var t2' x y), l2')
        else
          let (t1', l1') = (cf_term' t1 (x ++ l)) in
          let (t2', l2') = (cf_term' t2 l1') in
            (Forall (x, t1', t2'), l2')
    | Phi (x, n, t1, t2) ->
        if (VarSet.subset (VarSet.singleton x) l) then
          let y = (fresh_name_list_name l x) in
          let (t11_concl, t11_lcv) = (forall_nth_concl_list_var_type t1 n) in
          let (t',l') = (
                         (
                           cf_term' 
                             (
                               concl_list_var_type_forall 
                                 (App (t11_concl, t2))
                                 t11_lcv
                             )
                         )
                         (y ++ l)
          ) in
          let (t'_concl, t'_lcv) = (forall_nth_concl_list_var_type t' n) in (
            match t'_concl with
              | App (t1', t2') -> (
                  rewrite_term_var_var
                  
                    (Phi (
                      y,
                      n,
                      concl_list_var_type_forall t1' t'_lcv,
                      t2'
                    ))                   
                    x 
                    y
                    
                    ,l')
                    
                
              | _ -> (t,l)
          )
        else
          let (t11_concl, t11_lcv) = (forall_nth_concl_list_var_type t1 n) in
          let (t',l') = (cf_term' 
                         (
                             (
                               concl_list_var_type_forall 
                                 (App (t11_concl, t2))
                                 t11_lcv
                             )
                         )
                         (x ++ l)
          ) in
          let (t'_concl, t'_lcv) = (forall_nth_concl_list_var_type t' n) in (
            match t'_concl with
              | App (t1', t2') -> (
                  
                  (
                    Phi (
                      x,
                      n,
                      concl_list_var_type_forall t1' t'_lcv,
                      t2'
                    )                   
                    
                      ,l')
                    
                )
              | _ -> (t,l)
          )

    | Ind (x, t1, t2) ->
        if (VarSet.subset (VarSet.singleton x) l) then
          let y = (fresh_name_list_name l x) in
          let (t1', l1') = (cf_term' t1 (y ++ l)) in
          let (t2', l2') = (cf_listterm' t2 l1') in
            (Ind (y, rewrite_term_var_var t1' x y, rewrite_listterm_var_var t2' x y), l2')
        else
          let (t1', l1') = (cf_term' t1 (x ++ l)) in
          let (t2', l2') = (cf_listterm' t2 l1') in
            (Ind (x, t1', t2'), l2')
    | App (t1, t2) ->
        let (t1', l1') = (cf_term' t1 l) in
        let (t2', l2') = (cf_term' t2 l1') in
          (App (t1',t2'), l2')
    | Match (t1, l1) ->
        let (t1', l1') = (cf_term' t1 l) in
        let (t2', l2') = (cf_listterm' l1 l1') in
          (Match (t1', t2'), l2')
    | MatchT (t1, l1, t2) ->
        let (t1', l1') = (cf_term' t1 l) in
        let (t2', l2') = (cf_listterm' l1 l1') in
        let (t3', l3') = (cf_term' t2 l2') in
          (MatchT (t1', t2', t3'), l3')
    | Cons (n, l1) ->
        let (t1', l1') = (cf_term' l1 l) in
          (Cons (n, t1'), l1')
    | Var x -> (Var x, l)
    | _ -> (t, l)
and cf_listterm' l lv =
  match l with
    | [] -> 
        ([], lv)
    | hd :: tl ->
        let (l', lv') = (cf_term' hd lv) in 
        let (l'', lv'') = (cf_listterm' tl lv') in
          (l'::l'', lv'');;

let rec cf_context' c l =
  match c with
    | [] -> (c,l)
    | (hd1,hd2) :: tl -> (
        match (cf_term' hd2 (hd1 ++ l)) with
          | (hd2', l') -> (
              match (cf_context' tl l') with
                | (tl', l'') ->
                    ((hd1,hd2')::tl', l')
            )
      );;

let cf_term t = 
  let (t', l') = (cf_term' t VarSet.empty) in t';;

let cf_term_listvar t l =
  let (t', l') = (cf_term' t l) in t';;

let rec term_size t =
  match t with
    | Var v -> 
        1
    | Lambda (s, t1, t2) ->
         1 + term_size t1 + term_size t2
    | Forall (s, t1, t2) ->
        1 + term_size t1 + term_size t2
    | Phi (s, n, t1, t2) ->
        1 + term_size t1 + term_size t2
    | Ind (s, t1, l) ->
        1 + term_size t1 + listterm_size l
    | App (t1, t2) ->
        1 + term_size t1 + term_size t2
    | Match (t1, l) ->
        1 + term_size t1 + listterm_size l
    | MatchT (t1, l, t2) ->
        1 + term_size t1 + listterm_size l
    | Cons (n, t1) ->
        1 + term_size t1
    | Type -> 1
and listterm_size l =
  match l with
    | [] -> 0
    | hd :: tl ->
        term_size hd + listterm_size tl;;

let rec context_size c =
  match c with
    | [] -> 0
    | (hd1, hd2) :: tl ->
        term_size hd2 + context_size tl;;
        

let subst_var_term t x t' bv =
  rewrite_term_var_term 
    (cf_term_listvar t (x ++ (bv +++ (term_fv t' bv))))
    x
    t';;

let rec context_set_var c =
  match c with
    | [] -> VarSet.empty
    | (hd1,hd2)::tl ->
        hd1 ++ (context_set_var tl);;
