(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Def;;
open Printf;;
open Freshness;;
open Pprinter;;
open Listutil;;
open Varset;;
open Term;;
open Termequiv;;
open Unfold;;

(* composition of substitution *)

let rec apply_subs_subs_rev x t s = 
  match s with
    | [] -> []
    | (x', t') :: tl ->
        (x', apply_subs_term t' ((x,t) :: []) VarSet.empty) :: (apply_subs_subs_rev x t tl);;

let rec comp_subs_rev s =
  match s with
    | [] -> []
    | (x, t) :: tl ->
        (x, t) :: comp_subs_rev (apply_subs_subs_rev x t tl);;

let comp_subs s = rev (comp_subs_rev (rev s));;
 
        


(* application of substitution *)
let rec apply_subst_var x s =
  match s with
    | [] -> Var x
    | (hd1,hd2) :: tl ->
        if (hd1 = x) then
          hd2 else
          apply_subst_var x tl;;

(* pas tout a fait correct, devrait include un set de var quantifier nan ? *)
let rec apply_subst_term t s sv =
  match t with
    | Var v -> 
        if (VarSet.subset (VarSet.singleton v) sv) then
          Var v else
          apply_subst_var v s
    | Lambda (x, t1, t2) ->
        Lambda (x, apply_subst_term t1 s sv, apply_subst_term t2 s (x ++ sv))
    | Forall (x, t1, t2) ->
        Forall (x, apply_subst_term t1 s sv, apply_subst_term t2 s (x ++ sv))
    | Phi (x, n, t1, t2) ->
        Phi (x, n, apply_subst_term t1 s sv, apply_subst_term t2 s (x ++ sv))
    | Ind (x, t1, l) ->
        Ind (x, apply_subst_term t1 s sv, apply_subst_listterm l s (x ++ sv))
    | App (t1, t2) ->
        App (apply_subst_term t1 s sv, apply_subst_term t2 s sv)
    | Match (t1, l) ->
        Match (apply_subst_term t1 s sv, apply_subst_listterm l s sv)
    | MatchT (t1, l, t2) ->
        MatchT (apply_subst_term t1 s sv, apply_subst_listterm l s sv, apply_subst_term t2 s sv)
    | Cons (n, t1) ->
        Cons (n, apply_subst_term t1 s sv)
    | Type -> Type
and apply_subst_listterm l s sv =
  match l with
    | [] -> []
    | hd :: tl ->
        (apply_subst_term hd s sv) :: (apply_subst_listterm tl s sv);;
  
let rec apply_subst_context c s sv =
  match c with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, apply_subst_term hd2 s sv) :: (apply_subst_context tl s sv);;

(* unification (but not under quantification) *)

let rec unification t1 t2 bv =
  (*printf "[ "; print_term t1; printf " | "; print_term t2; printf " ] (";
  print_list bv; printf")\n\n";*)
  match (t1,t2) with
    | (Var x1, Var _) ->
        Mgu ((x1, t2)::[])
    | (Var x1, _) ->
        (*if (VarSet.subset (VarSet.singleton x1) bv) then (Some []) else*)
        if (VarSet.subset (VarSet.singleton x1) (term_fv t2 bv)) then
          NoMgu else
          Mgu ((x1, t2)::[])
    | (_, Var x2) ->
        (*if (VarSet.subset (VarSet.singleton x2) bv) then (Some []) else*)
        if (VarSet.subset (VarSet.singleton x2) (term_fv t1 bv)) then
          NoMgu else
          Mgu ((x2, t1)::[])
    | (Cons (n0, t1), Cons (n1, t2)) ->
        if (n0 = n1 && term_equiv t1 t2 bv []) then
          Mgu [] else NoMgu
          
    | (App(_,_), _) ->
        let (f1, a1) = (app_first_term_args t1) in 
        let (f2, a2) = (app_first_term_args t2) in (
          match (f1, f2) with
(*            | (Phi(_,_,_,_), _) -> DnkMgu
            | (_, Phi(_,_,_,_)) -> DnkMgu
            | (Lambda(_,_,_), _) -> DnkMgu
            | (_, Lambda(_,_,_)) -> DnkMgu
            | (Forall(_,_,_), _) -> DnkMgu
            | (_, Forall(_,_,_)) -> DnkMgu*)
            | (_, _) ->
                unification_list (f1::a1) (f2::a2) bv
        )

    | (_ , App(_,_)) ->
        let (f1, a1) = (app_first_term_args t1) in 
        let (f2, a2) = (app_first_term_args t2) in (
          match (f1, f2) with
(*            | (Phi(_,_,_,_), _) -> DnkMgu
            | (_, Phi(_,_,_,_)) -> DnkMgu
            | (Lambda(_,_,_), _) -> DnkMgu
            | (_, Lambda(_,_,_)) -> DnkMgu
            | (Forall(_,_,_), _) -> DnkMgu
            | (_, Forall(_,_,_)) -> DnkMgu*)
            | (_, _) ->
                unification_list (f1::a1) (f2::a2) bv
        )

    | _ -> (if (term_equiv t1 t2 bv []) then (Mgu []) else (DnkMgu))

and unification_list l1 l2 bv =
  match (l1, l2) with
    | ([],[]) -> Mgu []
    | (hd1::tl1, hd2::tl2) ->(
        match (unification hd1 hd2 bv) with
          | NoMgu -> NoMgu
          | DnkMgu -> DnkMgu
          | Mgu s ->
              match (unification_list (apply_subst_listterm tl1 s bv) (apply_subst_listterm tl2 s bv) bv) with
                | NoMgu -> NoMgu
                | DnkMgu -> DnkMgu
                | Mgu s' -> Mgu (s @ s')
      )
    | _ -> NoMgu;;


let rec left_unification t1 t2 bv =
  (*printf "[ "; print_term t1; printf " | "; print_term t2; printf " ] (";
  print_list bv; printf")\n\n";*)
  match (t1,t2) with
    | (Var x1, _) ->
        (*if (VarSet.subset (VarSet.singleton x1) bv) then (Some []) else*)
        (*if (VarSet.subset (VarSet.singleton x1) (term_fv t2 bv)) then
            None else*)
              Mgu ((x1, t2)::[])
    | (App (t11, t12), App(t21,t22)) -> (
        match (left_unification t11 t21 bv) with
          | NoMgu -> NoMgu
          | DnkMgu -> DnkMgu
          | Mgu s1 -> (
              match (left_unification (apply_subs_term t12 s1 bv) t22 bv) with
                | NoMgu -> NoMgu
                | DnkMgu -> DnkMgu
                | Mgu s2 ->
                    Mgu (s1 @ s2)
            )
      )

    | _ -> if (term_equiv t1 t2 bv []) then Mgu [] else DnkMgu

let rec left_unification_list l1 l2 bv =
  match (l1, l2) with
    | ([],[]) -> Mgu []
    | (hd1::tl1, hd2::tl2) ->(
        match (left_unification hd1 hd2 bv) with
          | NoMgu -> NoMgu
          | DnkMgu -> DnkMgu
          | Mgu s ->
              match (left_unification_list (apply_subst_listterm tl1 s bv) tl2 bv) with
                | NoMgu -> NoMgu
                | DnkMgu -> DnkMgu
                | Mgu s' -> Mgu (s @ s')
      )
    | _ -> NoMgu;;


let rec match_var_args x from into =
  match (from, into) with
    | ((hd11, hd12) :: tl1, (hd21, hd22) :: tl2) ->
        if (x = hd11) then hd21 else
          (match_var_args x tl1 tl2)
    | _ -> x;;

let rec match_comp_args s from into =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (match_var_args hd1 from into, hd2) :: (match_comp_args tl from into);;

let rec match_var_args2 t from into =
  match (from, into) with
    | ((hd11, hd12) :: tl1, (hd21, hd22) :: tl2) ->
        match_var_args2 (rewrite_term_var_term t hd11 (Var hd21)) tl1 tl2
    | _ -> t;;

let rec match_comp_args2 s from into =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (match_var_args hd1 from into, match_var_args2 hd2 from into) :: (match_comp_args2 tl from into);;

let rec sub_mult_subs x t s =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        if (VarSet.subset (VarSet.singleton hd1) (term_fv t VarSet.empty)) then
          (x, apply_subst_term t s VarSet.empty)::[] else (sub_mult_subs x t tl);;

let rec subs_mult_subs s1 s2 =
  match s1 with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (sub_mult_subs hd1 hd2 s2) @ (subs_mult_subs tl s2);;


let rec subs_cste s =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        let tl' = subs_cste tl in
          (*printf "subs_cste : \n";
          print_term hd2;
          printf "\n";
          print_list (VarSet.elements (term_fv hd2 VarSet.empty));
          printf "\n\n";*)
          if (term_fv hd2 VarSet.empty = VarSet.empty) then 
            (hd1, hd2) :: tl' else
              tl';;
          
